

import sys
import tempfile
import cPickle

def get_class( kls ):
    """
    Function returns a class given its qualified name i.e. <module_name>.<class_name>
    """
    parts = kls.split('.')
    module = ".".join(parts[:-1])
    m = __import__( module )
    for comp in parts[1:]:
        m = getattr(m, comp)            
    return m

import inspect

def get_model_names(module):
    return [m[0] for m in inspect.getmembers(module, lambda c: inspect.isclass(c) and c.__module__=='models_auxin_final')]


from import_tissue import import_tissue
from openalea.scheduler import Scheduler,Task
from parameters import setSimParameters
from db_utilities import get_parameters, set_parameters, get_graph, get_cparameter, set_cparameter
import models_auxin_final as models
import cellset_to_oa

from genenetwork import GeneNetworkJPat, CombinedModel
import os
from model_output import CellSetOutput, OutputFile, OutputDirectory
from manipulatePINandAUX1_revised2 import manipulate

def eid_to_cid_wid(db):
    """
    Returns a map going from edge ids to (cell, wall) ids
    """
    graph = get_graph(db)
    wall = db.get_property('wall')
    eid_map = {}
    for eid in graph.edges():
        eid_map[eid] = (int(graph.source(eid)), int(wall[eid]))
    return eid_map

def cid_wid_to_eid(db):
    """
    """
    eid_map = eid_to_cid_wid(db)
    return dict((v,k) for (k,v) in eid_map.iteritems())

from oa_to_xml import write_xml, get_xml_string

class Simulation(object):
    def __init__(self):
        self.models = []
        self.modules = []
        self.sch = Scheduler()
        self.steps = 0
        self.sim_initialized = False
        self.param_dict = {}
        self.geometry_changed_flag = False

    def initialize_tissue(self):
        """
        Imports a tissue and sets the simulation parameters
        """
        self.db = import_tissue("./geometries/" + self.tissue_name)
        self.db.set_property('parameters', self.param_dict)
        self.db.set_description('parameters', 'Model Parameters')
        setSimParameters(self.db)
        self.modules = []
        self.sim_initialized = False
        return True

    def set_tissue_name(self, filename):
        """
        """
        self.tissue_name = filename
	return self.tissue_name

    def add_model(self, model, descr=None):
        self.models.append((model, descr))
        model.set_default_parameters(self.param_dict)
        return True

    def add_model_by_name(self, name):
        if name not in [m[1] for m in self.models]:
            m = getattr(models, name)
            return self.add_model(m(), name)
        else:
            return True


    def list_available_tissuefiles(self):
        """
	Returns a list of available tissue files (zip format)
	"""
        l = list()
	import glob
	import os
	os.chdir("./geometries/")
	for files in glob.glob("*.zip"):
	    l.append(files)
	os.chdir("../")
 	return l

    def list_available_models(self):
	"""
  	Returns a list of available models
	"""
        return get_model_names(models)

    def list_loaded_models(self):
        """
`	Returns a list of models already loaded in the simulation
	"""
        return [m[1] for m in self.models]

    def get_model_default_parameters(self, name):
	"""
	Returns a dictionary of model parameters
	"""
        if name in self.list_loaded_models():
            i = self.list_loaded_models().index(name)
            return self.models[i][0].default_parameters()
        else:
            m = getattr(models, name)
            return m().default_parameters()

    def get_species_annotation(self, species, key):
        if key=='realname':
            try:
                return models.species_realnames.get(species, species)
            except:
                return species
        else:
            return ''

    def get_model_species_names(self, name):
	"""
	Returns a list of cell species names specified by a model
	"""
        if name in self.list_loaded_models():
            i = self.list_loaded_models().index(name)
            l = dict(self.models[i][0].get_species() + self.models[i][0].get_modifier_species())
        else:
            m = getattr(models, name)
            l = dict(m().get_species() + m().get_modifier_species())
        l = dict((k,v) if v!='edge' else (k,'(cell,wall)') for k,v in l.iteritems())
        return l

    def get_loaded_species_names(self):
	"""
	Returns a dictionary of species names that have been loaded in the simulation
	"""
	try:
            dp = self.db.get_property('divided_props')
    	    species = [ (s, dp[s][0]) for s in dp if dp[s][1]!='attribute' ]
            return dict((k,v) if v!='edge' else (k,'(cell,wall)') for k,v in species)
	except:
	    return dict()

    def get_species_data(self, name):
	try:
            prop_type = self.db.get_property('divided_props')[name][0]
            if prop_type != 'edge':
                return self.db.get_property(name)
            else:
               prop = self.db.get_property(name)
               eid_map = eid_to_cid_wid(self.db)
               return dict(('(%d_%d)'%eid_map[eid], v) for eid,v in prop.iteritems()) 
        except:
	    return dict()
        
    def set_cell_species_data(self, cell_id, species_name, value):
        try:
            set_species_data(species_name, dict(cell_id = value))
            return True
        except:
            return False

    def set_species_data(self, species_name, species_data, count_number=False):
        try:
            prop_type = self.db.get_property('divided_props')[species_name][0]
            if prop_type != 'edge':
                prop = self.db.get_property(species_name)
                if count_number:
                    c=sum(1 for k in species_data if k in prop)
                prop.update(species_data)
                if count_number:
                    return c
                else:
                    return True
            else:
                prop = self.db.get_property(species_name)
                eid_map = cid_wid_to_eid(self.db)
                if count_number:
                    c=sum(1 for k in species_data if eid_map[k] in prop)
                prop.update((eid_map[k], v) for k,v in species_data.iteritems())
                if count_number:
                    return c
                else:
                    return True
        except:
            return False

        
    def init_simulation(self):
        if not self.sim_initialized:
            self.modules=[]
       	    self.combined_model = CombinedModel([m[0] for m in self.models])
            gn = GeneNetworkJPat(self.db, self.combined_model)
            self.modules.append(gn)
            self.sim_initialized=True
        return self.sim_initialized

    def step(self):
	self.init_simulation()
        for m in self.modules:
            m.step()
        self.steps += 1
        return self.steps

    def set_iteration(self, value):
	self.steps = int(value)
	if self.steps == 0:
            self.sim_initialized = False;
        return self.steps

    def get_iteration(self):
        """
	Returns the current iteration number
	"""
        return self.steps

    def set_parameter(self, param, value):
        set_cparameter(self.param_dict, param, value)
        return get_cparameter(self.param_dict, param)


    def get_parameter(self, param):
        return get_cparameter(self.param_dict, param)


    def get_geometry(self):
	"""
	Returns an XML string (CellSet format) representing the tissue geometry
	"""
        self.geometry_changed_flag = False
        return get_xml_string(self.db, preserve_oa_ids=True)

    def set_geometry(self, xmlGeometry):
        pass
        
    def geometry_changed(self):
        return self.geometry_changed_flag

    def reset_simulation(self):
	"""
 	Resets all simulation data, bar the tissue geometry
        and the models ....
	"""
        self.param_dict={}
        for m in self.models:
            m[0].set_default_parameters(self.param_dict)
        
	return False
        
    def set_carrier_on_celltype(self, dictValues):
        """
        Sets AUX1/PIN carrier levels to a specified @value, on all cells of a specified type. 
        """
	return manipulate(self.db, dictValues)

    def dump_state(self, filename):
        cPickle.dump(self, open(filename,'w'))
        return True

    def save_tissue(self, filename):
        self.db.write(filename)
        return True

        
